@@ -32,9 +32,10 @@ namespace logmsg
 		debug_verbose = 1ull << 6,
 		debug_debug   = 1ull << 7,
 
-
-		private1     = 1ull << 32,
-		private32    = 1ull << 63
+		/// The types in the range custom1 to custom32 are free to use
+		/// by programs using libfilezilla for their own use
+		custom1     = 1ull << 32,
+		custom32    = 1ull << 63
 	};
 }
 
@@ -46,10 +47,10 @@ namespace logmsg
  * The actual string to log gets assembled from the format string and its
  * arguments only if the type is supposed to be logged.
  */
-class logger_interface
+class FZ_PUBLIC_SYMBOL logger_interface
 {
 public:
-	logger_interface() = default;
+	logger_interface() noexcept = default;
 	virtual ~logger_interface() = default;
 
 	logger_interface(logger_interface const&) = delete;
@@ -59,12 +60,30 @@ public:
 	/// The one thing you need to override
 	virtual void do_log(logmsg::type t, std::wstring && msg) = 0;
 
-	/// The \arg fmt argument is a format string suitable for fz::sprintf
+	/**
+	 * The \arg fmt argument is a format string suitable for fz::sprintf
+	 *
+	 * Assumes that all narrow string arguments are in the locale's current encoding.
+	 */
 	template<typename String, typename...Args>
 	void log(logmsg::type t, String&& fmt, Args&& ...args)
 	{
 		if (should_log(t)) {
-			std::wstring formatted = fz::to_wstring(fz::sprintf(std::forward<String>(fmt), std::forward<Args>(args)...));
+			std::wstring formatted = fz::sprintf(fz::to_wstring(std::forward<String>(fmt)), args...);
+			do_log(t, std::move(formatted));
+		}
+	}
+
+	/**
+	 * The \arg fmt argument is a format string suitable for fz::sprintf
+	 *
+	 * Assumes that all narrow string arguments, excluding the format string, are in UTF-8
+	 */
+	template<typename String, typename...Args>
+	void log_u(logmsg::type t, String&& fmt, Args const& ...args)
+	{
+		if (should_log(t)) {
+			std::wstring formatted = fz::sprintf(fz::to_wstring(std::forward<String>(fmt)), assume_strings_are_utf8(args)...);
 			do_log(t, std::move(formatted));
 		}
 	}
@@ -116,8 +135,48 @@ public:
 
 protected:
 	std::atomic<uint64_t> level_{logmsg::status | logmsg::error | logmsg::command | logmsg::reply};
+
+private:
+	std::wstring assume_strings_are_utf8(std::string_view const& arg) {
+		return fz::to_wstring_from_utf8(arg);
+	}
+	std::wstring assume_strings_are_utf8(std::string const& arg) {
+		return fz::to_wstring_from_utf8(arg);
+	}
+	std::wstring assume_strings_are_utf8(char const* arg) {
+		return fz::to_wstring_from_utf8(arg);
+	}
+
+	template <typename T>
+	T const& assume_strings_are_utf8(T const& arg) {
+		return arg;
+	}
 };
+
+/// A logger that does not log anything.
+class FZ_PUBLIC_SYMBOL null_logger final : public logger_interface
+{
+public:
+	null_logger() noexcept
+	{
+		disable(levels());
+	}
+
+	virtual void do_log(logmsg::type, std::wstring &&) override {}
+	virtual void set_all(logmsg::type) override {}
+	virtual void enable(logmsg::type) override {}
+};
+
+null_logger FZ_PUBLIC_SYMBOL & get_null_logger();
+
+/// A simple logger that writes to stdout
+class FZ_PUBLIC_SYMBOL stdout_logger final : public logger_interface
+{
+public:
+	virtual void do_log(logmsg::type, std::wstring &&) override;
+};
+
+
 }
 
 #endif
-
