@@ -23,6 +23,30 @@ struct certificate_verification_event_type;
  */
 typedef simple_event<certificate_verification_event_type, tls_layer*, tls_session_info> certificate_verification_event;
 
+enum class tls_ver
+{
+	v1_0,
+	v1_1,
+	v1_2,
+	v1_3
+};
+
+enum class tls_server_flags : unsigned int
+{
+	none = 0,
+
+	/// In TLS 1.3, do not automatically send PSKs after finishing handshake. Ignored if not TLS 1.3
+	no_auto_ticket = 0x1
+};
+
+inline bool operator&(tls_server_flags lhs, tls_server_flags rhs) {
+	return (static_cast<std::underlying_type_t<tls_server_flags>>(lhs) & static_cast<std::underlying_type_t<tls_server_flags>>(rhs)) != 0;
+}
+inline tls_server_flags operator|(tls_server_flags lhs, tls_server_flags rhs) {
+	return static_cast<tls_server_flags>(static_cast<std::underlying_type_t<tls_server_flags>>(lhs) | static_cast<std::underlying_type_t<tls_server_flags>>(rhs));
+}
+
+
 /**
  * \brief A Transport Layer Security (TLS) layer
  *
@@ -81,20 +105,13 @@ public:
 	 *
 	 * Before calling server_handshake, a valid certificate and key must be passed
 	 * in through \ref set_certificate.
+	 *
+	 * Session parameters of an existing session can be passed to allow session resumption. Check after handshake completion
+	 * with resumed_session()
+	 *
+	 * The preamble is sent out after setting up all the parameters, but before the first handshake message
 	 */
-	bool server_handshake();
-
-	/**
-	 * \brief Starts shaking hand for a new TLS session as server.
-	 *
-	 * Returns true if the handshake has started, false on error.
-	 *
-	 * If the handshake is started, wait for a connection event for the result.
-	 *
-	 * Before calling server_handshake, a valid certificate and key must be passed
-	 * in through \ref set_certificate.
-	 */
-	bool server_handshake(std::vector<uint8_t> const& session_to_resume);
+	bool server_handshake(std::vector<uint8_t> const& session_to_resume = {}, std::string_view const& preamble = {}, tls_server_flags flags = {});
 
 	/// Gets session parameters for resumption
 	std::vector<uint8_t> get_session_parameters() const;
@@ -140,7 +157,7 @@ public:
 	 *
 	 * If the pem flag is set, the input is assumed to be in PEM, otherwise DER.
 	 */
-	bool set_certificate(std::string const& key, std::string const& certs, native_string const& password, bool pem = true);
+	bool set_certificate(std::string_view const& key, std::string_view const& certs, native_string const& password, bool pem = true);
 
 	/// Returns the version of the loaded GnuTLS library, may be different than the version used at compile-time.
 	static std::string get_gnutls_version();
@@ -154,7 +171,66 @@ public:
 	 * The output pair is in PEM, first element is the key and the second the certificate.
 	 */
 	static std::pair<std::string, std::string> generate_selfsigned_certificate(native_string const& password, std::string const& distinguished_name, std::vector<std::string> const& hostnames);
+	static std::pair<std::string, std::string> generate_csr(native_string const& password, std::string const& distinguished_name, std::vector<std::string> const& hostnames, bool csr_as_pem = true);
 
+	/** \brief Negotiate application protocol
+	 *
+	 * If the peer makes use of ALPN, the handshake fails if no matching protocol is found.
+	 * If the peer does not use/support ALPN, the handshake continues and no protocol is negotiated.
+	 * 
+	 * Needs to be called prior to handshaking.
+	 *
+	 * Which protocol is eventually chosen is always a server-side decsion.
+	 * Iff running as server, server_priority controls the ALPN should be chosen
+	 * based on the server's or the client's order in the list of ALPNs supported
+	 * by both. server_priority is ignored if unning as client.
+	 */
+	bool set_alpn(std::string_view const& alpn);
+	bool set_alpn(std::vector<std::string> const& alpns, bool server_priority = false);
+
+	/** \brief Sets minimum allowed TLS version
+	 */
+	void set_min_tls_ver(tls_ver ver);
+
+	/** \brief Sets maximum allowed TLS versions
+	 *
+	 * Don't set a max version in production, it is for testing things.
+	 */
+	void set_max_tls_ver(tls_ver ver);
+
+	/// After a successful handshake, returns which protocol, if any, has been negotiated
+	std::string get_alpn() const;
+
+	/// If running as server, get the SNI sent by the client
+	native_string get_hostname() const;
+
+	bool is_server() const;
+
+	/** \brief If running as server with TLS1.3, send out a new session ticket before the next data payload.
+	 *
+	 * Returns 0 on success, socket error otheerwise. Never returns EAGAIN.
+	 *
+	 * It's a NOOP if TLS version != 1.3
+	 *
+	 * Returns EINVAL if not server.
+	 */
+	int new_session_ticket();
+
+	/** \brief Sets a callback to control whether unexpected eof is seen as error
+	 *
+	 * With TLS, an EOF on the socket prior to receiving a closure alert normally is an error.
+	 * In many cases this is harmless though, e.g. if the connection is idle.
+	 *
+	 * If this callback is set, premature termination is no longer seen as error if the
+	 * callback returns false.
+	 *
+	 * Callback must not call any tls_layer function.
+	 *
+	 * Only controls the layer's own logging. Functions such as read will still return
+	 * ECONNABORTED.
+	 */
+	void set_unexpected_eof_cb(std::function<bool()> const& cb);
+	void set_unexpected_eof_cb(std::function<bool()> && cb);
 
 	virtual socket_state get_state() const override;
 
@@ -167,6 +243,8 @@ public:
 
 	virtual int shutdown_read() override;
 
+	virtual void set_event_handler(event_handler* pEvtHandler, fz::socket_event_flag retrigger_block = socket_event_flag{}) override;
+
 private:
 	virtual void FZ_PRIVATE_SYMBOL operator()(event_base const& ev) override;
 
