@@ -2,12 +2,12 @@
 #include "libfilezilla/impersonation.hpp"
 #include "libfilezilla/process.hpp"
 #include "libfilezilla/thread_pool.hpp"
+#include "libfilezilla/util.hpp"
 
 #ifdef FZ_WINDOWS
 
 #include "libfilezilla/buffer.hpp"
 #include "libfilezilla/encode.hpp"
-#include "libfilezilla/util.hpp"
 #include "libfilezilla/glue/windows.hpp"
 #include "windows/security_descriptor_builder.hpp"
 
@@ -382,7 +382,19 @@ public:
 		handler_->event_loop_.filter_events(process_event_filter);
 	}
 
-	bool kill(bool wait = true, bool force = false)
+	template<typename Out, typename In>
+	constexpr Out clamped_cast(In in) noexcept
+	{
+		if (cmp_less(in, std::numeric_limits<Out>::min())) {
+			return std::numeric_limits<Out>::min();
+		}
+		if (cmp_less(std::numeric_limits<Out>::max(), in)) {
+			return std::numeric_limits<Out>::max();
+		}
+		return static_cast<Out>(in);
+	}
+
+	bool kill(bool force = true, duration const& timeout = {})
 	{
 		if (handler_) {
 			{
@@ -423,7 +435,8 @@ public:
 				TerminateProcess(process_handle_, 0);
 			}
 			else {
-				if (WaitForSingleObject(process_handle_, wait ? INFINITE : 0) == WAIT_TIMEOUT) {
+				DWORD wait = (timeout >= duration()) ? clamped_cast<DWORD>(timeout.get_milliseconds()) : INFINITE;
+				if (WaitForSingleObject(process_handle_, wait) == WAIT_TIMEOUT) {
 					return false;
 				}
 			}
@@ -461,18 +474,6 @@ public:
 		}
 	}
 
-	template<typename Out, typename In>
-	constexpr Out clamped_cast(In in) noexcept
-	{
-		if (cmp_less(in, std::numeric_limits<Out>::min())) {
-			return std::numeric_limits<Out>::min();
-		}
-		if (cmp_less(std::numeric_limits<Out>::max(), in)) {
-			return std::numeric_limits<Out>::max();
-		}
-		return static_cast<Out>(in);
-	}
-
 	rwresult read(void* buffer, size_t len)
 	{
 		if (!len || out_.read_ == INVALID_HANDLE_VALUE) {
@@ -579,9 +580,9 @@ private:
 	event_handler * handler_{};
 
 	mutex mutex_;
-	fz::async_task task_;
-	fz::buffer read_buffer_;
-	fz::buffer write_buffer_;
+	async_task task_;
+	buffer read_buffer_;
+	buffer write_buffer_;
 	HANDLE sync_{INVALID_HANDLE_VALUE};
 	OVERLAPPED ol_read_{};
 	OVERLAPPED ol_write_{};
@@ -606,6 +607,7 @@ private:
 #include <errno.h>
 #include <fcntl.h>
 #include <signal.h>
+#include <sys/syscall.h>
 #include <sys/wait.h>
 #include <string.h>
 #include <unistd.h>
@@ -902,7 +904,60 @@ public:
 		handler_->event_loop_.filter_events(process_event_filter);
 	}
 
-	bool kill(bool wait = true, bool force = false)
+	bool do_waitpid(bool wait = false)
+	{
+		pid_t ret;
+		do {
+			ret = waitpid(pid_, nullptr, wait ? 0 : WNOHANG);
+		} while (ret == -1 && errno == EINTR);
+		return ret != 0;
+	}
+
+	bool wait(duration const& timeout)
+	{
+		if (do_waitpid()) {
+			return true;
+		}
+		if (!timeout) {
+			return false;
+		}
+
+		monotonic_clock const deadline = monotonic_clock::now() + timeout;
+
+#ifdef SYS_pidfd_open
+		int pidfd = syscall(SYS_pidfd_open, pid_, 0);
+		if (pidfd >= 0) {
+			for (duration wait; wait = deadline - monotonic_clock::now(), wait > duration();) {
+				pollfd fds{};
+				fds.fd = pidfd;
+				fds.events = POLLIN;
+				if (poll(&fds, 1, wait.get_milliseconds()) > 0) {
+					if (do_waitpid()) {
+						close(pidfd);
+						return true;
+					}
+				}
+			}
+			close(pidfd);
+			return do_waitpid();
+		}
+#endif
+		// Sadly need to fall back to waidpid with WNOHANG.
+		// Poll fast at first, then slow down
+		//
+		// Signals are no option here, this is multi-threaded
+		// library code. We cannot control what other components are doing.
+		for (duration wait, sd = duration::from_milliseconds(1); wait = deadline - monotonic_clock::now(), wait > duration(); sd += sd / 10 + duration::from_milliseconds(1)) {
+			sleep(wait > sd ? sd : wait);
+			if (do_waitpid()) {
+				return true;
+			}
+		}
+
+		return false;
+	}
+
+	bool kill(bool force = true, duration const& timeout = {})
 	{
 		if (handler_) {
 			{
@@ -920,13 +975,13 @@ public:
 		if (pid_ != -1) {
 			::kill(pid_, force ? SIGKILL : SIGTERM);
 
-			pid_t ret;
-			do {
-				ret = waitpid(pid_, nullptr, wait ? 0 : WNOHANG);
-			} while (ret == -1 && errno == EINTR);
-
-			if (!ret) {
-				return false;
+			if (force || timeout < duration()) {
+				do_waitpid(true);
+			}
+			else {
+				if (!wait(timeout)) {
+					return false;
+				}
 			}
 
 			pid_ = -1;
@@ -1082,12 +1137,19 @@ bool process::spawn(std::vector<native_string> const& command_with_args, io_redi
 	return impl_ ? impl_->spawn(command_with_args.front(), begin, command_with_args.end(), redirect_mode) : false;
 }
 
-bool process::kill(bool wait, bool force)
+void process::kill()
 {
 	if (impl_) {
-		return impl_->kill(wait, force);
+		impl_->kill(true, {});
 	}
-	return true;
+}
+
+bool process::stop(duration const& timeout)
+{
+	if (impl_) {
+		return impl_->kill(false, timeout);
+	}
+	return false;
 }
 
 rwresult process::read(void* buffer, size_t len)
